#!/usr/bin/env python3
# coding: utf-8
import sys
from pathlib import Path
from shutil import make_archive
from tempfile import NamedTemporaryFile

from flask import (Blueprint, Response, abort, current_app, g, request,
                   send_file)
from flask.json import jsonify

from sapporo.const import GET_STATUS_CODE, POST_STATUS_CODE
from sapporo.run import (cancel_run, fork_run, get_run_log, prepare_run_dir,
                         validate_and_update_run_request, validate_run_id)
from sapporo.type import RunId, RunListResponse, RunLog, RunStatus, ServiceInfo
from sapporo.util import (generate_run_id, generate_service_info,
                          get_all_run_ids, get_run_dir, get_state,
                          path_hierarchy, secure_filepath, str2bool)

app_bp = Blueprint("sapporo", __name__)


@app_bp.route("/service-info", methods=["GET"])
def get_service_info() -> Response:
    """
    May include information related (but not limited to) the workflow
    descriptor formats, versions supported, the WES API versions supported,
    and information about general service availability.
    """
    res_body: ServiceInfo = generate_service_info()
    response: Response = jsonify(res_body)
    response.status_code = GET_STATUS_CODE

    return response


@app_bp.route("/runs", methods=["GET"])
def get_runs() -> Response:
    """
    This list should be provided in a stable ordering. (The actual ordering is
    implementation dependent.) When paging through the list, the client should
    not make assumptions about live updates, but should assume the contents of
    the list reflect the workflow list at the moment that the first page is
    requested. To monitor a specific workflow run, use GetRunStatus or
    GetRunLog.
    """
    if current_app.config["GET_RUNS"] is False:
        abort(403, "This endpoint `GET /runs` is unavailable because the "
              "service provider didn't allow the request to this endpoint "
              "when sapporo was started.")

    res_body: RunListResponse = {
        "runs": [],
        "next_page_token": ""
    }
    for run_id in get_all_run_ids():
        res_body["runs"].append({
            "run_id": run_id,
            "state": get_state(run_id).name  # type: ignore
        })
    response: Response = jsonify(res_body)
    response.status_code = GET_STATUS_CODE

    return response


@app_bp.route("/runs", methods=["POST"])
def post_runs() -> Response:
    """
    This endpoint creates a new workflow run and returns a `RunId` to monitor
    its progress.
    """
    run_id: str = generate_run_id()
    run_request = validate_and_update_run_request(
        run_id,
        dict(request.form),  # type: ignore
        request.files
    )
    prepare_run_dir(run_id, run_request, request.files)
    fork_run(run_id)
    response: Response = jsonify({
        "run_id": run_id
    })
    response.status_code = POST_STATUS_CODE

    return response


@app_bp.route("/runs/<string:run_id>", methods=["GET"])
def get_runs_id(run_id: str) -> Response:
    """
    This endpoint provides detailed information about a given workflow run.
    The returned result has information about the outputs produced by this
    workflow (if available), a log object which allows the stderr and stdout
    to be retrieved, a log array so stderr/stdout for individual tasks can be
    retrieved, and the overall state of the workflow run (e.g. RUNNING, see
    the State section).
    """
    validate_run_id(run_id)
    res_body: RunLog = get_run_log(run_id)
    response: Response = jsonify(res_body)
    response.status_code = GET_STATUS_CODE

    return response


@app_bp.route("/runs/<string:run_id>/cancel", methods=["POST"])
def post_runs_id_cancel(run_id: str) -> Response:
    """
    Cancel a running workflow.
    """
    validate_run_id(run_id)
    cancel_run(run_id)
    res_body: RunId = {"run_id": run_id}
    response: Response = jsonify(res_body)
    response.status_code = POST_STATUS_CODE

    return response


@app_bp.route("/runs/<string:run_id>/status", methods=["GET"])
def get_runs_id_status(run_id: str) -> Response:
    """
    This provides an abbreviated (and likely fast depending on implementation)
    status of the running workflow, returning a simple result with the overall
    state of the workflow run (e.g. RUNNING, see the State section).
    """
    validate_run_id(run_id)
    res_body: RunStatus = {
        "run_id": run_id,
        "state": get_state(run_id).name  # type: ignore
    }
    response: Response = jsonify(res_body)
    response.status_code = GET_STATUS_CODE

    return response


@app_bp.route("/runs/<string:run_id>/data/", methods=["GET"])
@app_bp.route("/runs/<string:run_id>/data/<path:subpath>", methods=["GET"])
def get_runs_id_data(run_id: str, subpath: str = "") -> Response:
    """
    This provides a remote url to download a file or directory under the
    `run_dir` of the Sapporo-service.

    - In the case of `path/to/file`, this returns the file.
    - In the case of `path/to/dir`, this returns the list of files under
      directory in JSON format.
    - In the case of `path/to/dir?download=true`, this returns the directory
      in zip format.

    The path is relative to the base directory of each run.
    See `README.md` in sapporo-service for the structure of `run_dir`.
    For example, if you want to download the output `foo.txt`, specify
    something like `outputs/foo.txt`.

    `..` will be ignored.
    """
    validate_run_id(run_id)
    requested_path = secure_filepath(subpath)
    path = get_run_dir(run_id).joinpath(secure_filepath(subpath))
    if not path.exists():
        parent = Path(f"runs/{run_id}/data").joinpath(requested_path.parent)
        abort(404,
              f"The specified path: {requested_path} does not exist. "
              f"Please make another request to `<endpoint>/{parent}/` again "
              "and check the dir structure.")
    if path.is_file():
        return send_file(path, as_attachment=True)
    else:
        if str2bool(request.args.get("download", False)):
            with NamedTemporaryFile() as f:
                res = make_archive(f.name, "zip",
                                   root_dir=path.parent, base_dir=path.name)
                if "temp_files" not in g:
                    g.temp_files = [Path(f"{f.name}.zip")]
                else:
                    g.temp_files.append(Path(f"{f.name}.zip"))
                return send_file(res, as_attachment=True,
                                 attachment_filename=f"{path.name}.zip")
        else:
            response: Response = jsonify(path_hierarchy(path, path))
            response.status_code = GET_STATUS_CODE
            return response


@app_bp.after_request
def delete_temp_files(response: Response) -> Response:
    if "temp_files" in g:
        for temp_file in g.temp_files:
            try:
                if sys.version_info.major == 3 and sys.version_info.minor >= 8:
                    temp_file.unlink(missing_ok=False)
                else:
                    temp_file.unlink()
            except Exception:
                pass

    return response



#!/usr/bin/env python3
# coding: utf-8
import collections
import json
import os
from pathlib import Path, PurePath
from typing import Any, Dict, Iterable, List, Union
from unicodedata import normalize
from uuid import uuid4

from flask import abort, current_app, request
from werkzeug.utils import _filename_ascii_strip_re  # type: ignore
from werkzeug.utils import _windows_device_files  # type: ignore

from sapporo.const import RUN_DIR_STRUCTURE
from sapporo.type import ServiceInfo, State, Workflow


def generate_service_info() -> ServiceInfo:
    with current_app.config["SERVICE_INFO"].open(mode="r") as f:
        service_info: ServiceInfo = json.load(f)

    service_info["supported_wes_versions"] = ["sapporo-wes-1.0.0"]
    service_info["system_state_counts"] = count_system_state()  # type: ignore
    service_info["tags"]["debug"] = current_app.config["DEBUG"]
    service_info["tags"]["get_runs"] = current_app.config["GET_RUNS"]
    service_info["tags"]["workflow_attachment"] = \
        current_app.config["WORKFLOW_ATTACHMENT"]
    service_info["tags"]["registered_only_mode"] = \
        current_app.config["REGISTERED_ONLY_MODE"]

    with current_app.config["EXECUTABLE_WORKFLOWS"].open(mode="r") as f:
        executable_workflows: List[Workflow] = json.load(f)
    service_info["executable_workflows"] = executable_workflows

    return service_info


def generate_run_id() -> str:
    return str(uuid4())


def get_run_dir(run_id: str) -> Path:
    run_base_dir: Path = current_app.config["RUN_DIR"]

    return run_base_dir.joinpath(run_id[:2]).joinpath(run_id).resolve()


def get_path(run_id: str, key: str) -> Path:
    run_dir: Path = get_run_dir(run_id)

    return run_dir.joinpath(RUN_DIR_STRUCTURE[key])


def get_all_run_ids() -> List[str]:
    run_base_dir: Path = current_app.config["RUN_DIR"]
    run_requests: List[Path] = \
        list(run_base_dir.glob(f"**/{RUN_DIR_STRUCTURE['run_request']}"))
    run_ids: List[str] = \
        [run_request.parent.name for run_request in run_requests]

    return run_ids


def get_state(run_id: str) -> State:
    try:
        with get_path(run_id, "state").open(mode="r") as f:
            str_state: str = \
                [line for line in f.read().splitlines() if line != ""][0]
        return State[str_state]
    except Exception:
        return State.UNKNOWN


def count_system_state() -> Dict[str, int]:
    run_ids: List[str] = get_all_run_ids()
    count: Dict[str, int] = \
        dict(collections.Counter(
            [get_state(run_id).name for run_id in run_ids]))

    return count


def write_file(run_id: str, file_type: str, content: str) -> None:
    file: Path = get_path(run_id, file_type)
    file.parent.mkdir(parents=True, exist_ok=True)
    with file.open(mode="w") as f:
        f.write(content)


def read_file(run_id: str, file_type: str) -> Any:
    json_file_type = ["run_request", "outputs", "wf_params"]
    oneline_txt_file_type = ["state", "start_time", "end_time",
                             "exit_code", "pid", "wf_engine_params", "cmd"]
    log_file_type = ["stdout", "stderr", "task_logs"]
    if file_type not in json_file_type + oneline_txt_file_type + log_file_type:
        return None
    file: Path = get_path(run_id, file_type)
    if file.exists() is False:
        return None
    with file.open(mode="r") as f:
        if file_type in json_file_type:
            return json.load(f)
        elif file_type in oneline_txt_file_type:
            return f.read().splitlines()[0]
        elif file_type in log_file_type:
            return f.read()


def dump_outputs_list(inputted_run_dir: str) -> None:
    run_dir: Path = Path(inputted_run_dir).resolve()
    run_id = run_dir.name
    with run_dir.joinpath(
            RUN_DIR_STRUCTURE["sapporo_config"]).open(mode="r") as f:
        sapporo_config = json.load(f)
    base_remote_url = \
        f"{sapporo_config['sapporo_endpoint']}/runs/{run_id}/data/"
    outdir_path: Path = run_dir.joinpath(RUN_DIR_STRUCTURE["outputs_dir"])
    output_files: List[Path] = sorted(list(walk_all_files(outdir_path)))
    outputs = []
    for output_file in output_files:
        outputs.append({
            "file_name": output_file.name,
            "file_url":
                f"{base_remote_url}{str(output_file.relative_to(run_dir))}"
        })
    with run_dir.joinpath(RUN_DIR_STRUCTURE["outputs"]).open(mode="w") as f:
        f.write(json.dumps(outputs, indent=2))


def walk_all_files(dir: Path) -> Iterable[Path]:
    for root, dirs, files in os.walk(dir):
        for file in files:
            yield Path(root).joinpath(file)


def get_workflow(workflow_name: str) -> Workflow:
    with current_app.config["EXECUTABLE_WORKFLOWS"].open(mode="r") as f:
        executable_workflows: List[Workflow] = json.load(f)
    for wf in executable_workflows:
        if wf["workflow_name"] == workflow_name:
            return wf

    abort(404,
          f"The workflow_name: {workflow_name} you requested doesn't "
          "exist. Please request `GET /service-info` again and check "
          "the registered executable workflows.")


def validate_wf_type(wf_type: str, wf_type_version: str) -> None:
    service_info: ServiceInfo = generate_service_info()
    wf_type_versions = service_info["workflow_type_versions"]

    available_wf_types: List[str] = list(map(str, wf_type_versions.keys()))
    if wf_type not in available_wf_types:
        abort(400,
              f"{wf_type}, the workflow_type specified in the "
              f"request, is not included in {available_wf_types}, "
              "the available workflow_types.")

    available_wf_versions: List[str] = \
        list(map(str, wf_type_versions[wf_type]["workflow_type_version"]))
    if wf_type_version not in available_wf_versions:
        abort(400,
              f"{wf_type_version}, the workflow_type_version specified in "
              f"the request, is not included in {available_wf_versions}, "
              "the available workflow_type_versions.")


def secure_filepath(filepath: str) -> Path:
    """
    We know `werkzeug.secure_filename()`.
    However, this function cannot represent the dir structure,

    >>> secure_filename("../../../etc/passwd")
    'etc_passwd'

    Thus, it is incompatible with workflow engines such as snakemake.
    Therefore, We implemented this by referring to `werkzeug.secure_filename()`

    Please check `tests/unit_test/test_secure_filepath.py`

    Reference of `PurePath.parts`:
    >> > PurePath("/").parts
    ('/',)
    >> > PurePath("//").parts
    ('//',)
    >> > PurePath("/foo/bar").parts
    ('/', 'foo', 'bar')
    >> > PurePath("foo/bar").parts
    ('foo', 'bar')
    >> > PurePath("/foo/bar/").parts
    ('/', 'foo', 'bar')
    >> > PurePath("./foo/bar/").parts
    ('foo', 'bar')
    >> > PurePath("/../../foo/bar//").parts
    ('/', '..', '..', 'foo', 'bar')
    >> > PurePath("/../.../foo/bar//").parts
    ('/', '..', '...', 'foo', 'bar')
    """
    ascii_filepath = \
        normalize("NFKD", filepath).encode("ascii", "ignore").decode("ascii")
    pure_path = PurePath(ascii_filepath)
    nodes = []
    for node in pure_path.parts:
        # Change space to underbar
        node = "_".join(node.split())
        # Change [^A-Za-z0-9_.-] to empty.
        node = str(_filename_ascii_strip_re.sub("", node))
        node = node.strip("._")
        if node not in ["", ".", ".."]:
            nodes.append(node)

    path = Path("/".join([str(node) for node in nodes]))

    if (
        os.name == "nt"
        and str(path)
        and str(path).split(".")[0].upper() in _windows_device_files
    ):
        path = Path("_" + str(path))

    return path


def path_hierarchy(original_path: Path, dir_path: Path) -> Any:
    hierarchy: Dict[str, Any] = {
        "type": "directory",
        "name": dir_path.name,
        "path": str(dir_path.relative_to(original_path)),
    }

    try:
        hierarchy["children"] = [
            path_hierarchy(original_path, dir_path.joinpath(child))
            for child in dir_path.iterdir()
        ]
    except Exception:
        hierarchy["type"] = "file"

    return hierarchy


def str2bool(val: Union[str, bool]) -> bool:
    if isinstance(val, bool):
        return val
    if val.lower() in ["true", "yes", "y"]:
        return True
    if val.lower() in ["false", "no", "n"]:
        return False

    return bool(val)


def dump_sapporo_config(run_id: str) -> str:
    host = request.host_url.strip("/")
    url_prefix = current_app.config['URL_PREFIX'].strip("/")
    endpoint = f"{host}/{url_prefix}".strip("/")

    return json.dumps({
        "get_runs": current_app.config["GET_RUNS"],
        "workflow_attachment": current_app.config["WORKFLOW_ATTACHMENT"],
        "registered_only_mode": current_app.config["REGISTERED_ONLY_MODE"],
        "service_info": str(current_app.config["SERVICE_INFO"]),
        "executable_workflows":
            str(current_app.config["EXECUTABLE_WORKFLOWS"]),
        "run_sh": str(current_app.config["RUN_SH"]),
        "url_prefix": current_app.config["URL_PREFIX"],
        "sapporo_endpoint": endpoint
    }, indent=2)



#!/usr/bin/env python3
# coding: utf-8
import argparse
import json
import os
import sys
from argparse import ArgumentParser, Namespace
from pathlib import Path
from traceback import format_exc
from typing import Dict, List, Optional, Union

from flask import Flask, Response, current_app, jsonify
from jsonschema import validate
from werkzeug.exceptions import HTTPException

from sapporo.const import (DEFAULT_ACCESS_CONTROL_ALLOW_ORIGIN,
                           DEFAULT_EXECUTABLE_WORKFLOWS, DEFAULT_HOST,
                           DEFAULT_PORT, DEFAULT_RUN_DIR, DEFAULT_RUN_SH,
                           DEFAULT_SERVICE_INFO, DEFAULT_URL_PREFIX,
                           EXECUTABLE_WORKFLOWS_SCHEMA, SERVICE_INFO_SCHEMA)
from sapporo.controller import app_bp
from sapporo.type import ErrorResponse, Workflow
from sapporo.util import str2bool


def parse_args(sys_args: List[str]) -> Namespace:
    parser: ArgumentParser = argparse.ArgumentParser(
        description="Implementation of a GA4GH workflow execution service "
                    "that can easily support various workflow runners.")

    parser.add_argument(
        "--host",
        nargs=1,
        metavar="",
        help=f"Host address of Flask. (default: {DEFAULT_HOST})"
    )
    parser.add_argument(
        "-p",
        "--port",
        type=int,
        nargs=1,
        metavar="",
        help=f"Port of Flask. (default: {DEFAULT_PORT})"
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Enable debug mode of Flask."
    )
    parser.add_argument(
        "-r",
        "--run-dir",
        nargs=1,
        metavar="",
        help="Specify the run dir. (default: ./run)"
    )
    parser.add_argument(
        "--disable-get-runs",
        action="store_true",
        help="Disable endpoint of `GET /runs`."
    )
    parser.add_argument(
        "--disable-workflow-attachment",
        action="store_true",
        help="Disable `workflow_attachment` on endpoint `Post /runs`."
    )
    parser.add_argument(
        "--run-only-registered-workflows",
        action="store_true",
        help="Run only registered workflows. Check the registered "
             "workflows using `GET /service-info`, and specify "
             "`workflow_name` in the `POST /run`."
    )
    parser.add_argument(
        "--service-info",
        nargs=1,
        metavar="",
        help="Specify `service-info.json`. The supported_wes_versions, "
             "system_state_counts and workflows are overwritten in the "
             "application."
    )
    parser.add_argument(
        "--executable-workflows",
        nargs=1,
        metavar="",
        help="Specify `executable-workflows.json`."
    )
    parser.add_argument(
        "--run-sh",
        nargs=1,
        metavar="",
        help="Specify `run.sh`."
    )
    parser.add_argument(
        "--url-prefix",
        nargs=1,
        metavar="",
        help="Specify the prefix of the url (e.g. --url-prefix /foo -> "
        "/foo/service-info)."
    )

    args: Namespace = parser.parse_args(sys_args)

    return args


def handle_default_params(args: Namespace) -> Dict[str, Union[str, int, Path]]:
    params: Dict[str, Union[str, int, Path]] = {
        "host": handle_default_host(args.host),
        "port": handle_default_port(args.port),
        "debug": handle_default_debug(args.debug),
        "run_dir": handle_default_path(args.run_dir,
                                       "SAPPORO_RUN_DIR",
                                       DEFAULT_RUN_DIR),
        "get_runs": handle_default_get_runs(args.disable_get_runs),
        "workflow_attachment":
            handle_default_workflow_attachment(
                args.disable_workflow_attachment),
        "registered_only_mode":
            handle_default_registered_only_mode(
                args.run_only_registered_workflows),
        "service_info": handle_default_path(args.service_info,
                                            "SAPPORO_SERVICE_INFO",
                                            DEFAULT_SERVICE_INFO),
        "executable_workflows":
            handle_default_path(args.executable_workflows,
                                "SAPPORO_EXECUTABLE_WORKFLOWS",
                                DEFAULT_EXECUTABLE_WORKFLOWS),
        "run_sh": handle_default_path(args.run_sh,
                                      "SAPPORO_RUN_SH",
                                      DEFAULT_RUN_SH),
        "url_prefix": handle_default_url_prefix(args.url_prefix)
    }

    return params


def handle_default_host(host: Optional[List[str]]) -> str:
    if host is None:
        return os.environ.get("SAPPORO_HOST", DEFAULT_HOST)

    return host[0]


def handle_default_port(port: Optional[List[str]]) -> int:
    if port is None:
        return int(os.environ.get("SAPPORO_PORT", DEFAULT_PORT))

    return int(port[0])


def handle_default_debug(debug: bool) -> bool:
    if debug is False:
        return str2bool(os.environ.get("SAPPORO_DEBUG", False))

    return True


def handle_default_path(input_arg: Optional[List[str]], env_var: str,
                        default_val: Path) -> Path:
    handled_path: Path
    if input_arg is None:
        handled_path = Path(os.environ.get(env_var, default_val))
    else:
        handled_path = Path(input_arg[0])
    if not handled_path.is_absolute():
        handled_path = Path.cwd().joinpath(handled_path).resolve()

    return handled_path


def handle_default_get_runs(disable_get_runs: bool) -> bool:
    if disable_get_runs:
        return False
    else:
        return str2bool(os.environ.get("SAPPORO_GET_RUNS", True))


def handle_default_workflow_attachment(disable_workflow_attachment: bool) \
        -> bool:
    if disable_workflow_attachment:
        return False
    else:
        return str2bool(os.environ.get("SAPPORO_WORKFLOW_ATTACHMENT", True))


def handle_default_registered_only_mode(run_only_registered_workflows: bool) \
        -> bool:
    if run_only_registered_workflows is False:
        return str2bool(os.environ.get("SAPPORO_RUN_ONLY_REGISTERED_WORKFLOWS",
                                       False))

    return True


def handle_default_url_prefix(url_prefix: Optional[List[str]]) -> str:
    if url_prefix is None:
        return os.environ.get("SAPPORO_URL_PREFIX", DEFAULT_URL_PREFIX)

    return url_prefix[0]


def check_uniqueness_wf_name(executable_wf_path: Path) -> None:
    with executable_wf_path.open(mode="r") as f:
        executable_wfs: List[Workflow] = json.load(f)
    wf_names: List[str] = [wf["workflow_name"] for wf in executable_wfs]
    if len(wf_names) != len(set(wf_names)):
        raise Exception(
            "`workflow_name` is not unique in the "
            f"`executable_workflows.json`: {executable_wf_path} you inputted.")


def validate_json(service_info: Path, executable_wf: Path) -> None:
    pairs: List[List[Path]] = [
        [service_info, SERVICE_INFO_SCHEMA],
        [executable_wf, EXECUTABLE_WORKFLOWS_SCHEMA]
    ]
    for data, schema in pairs:
        with data.open(mode="r") as f_d, schema.open(mode="r") as f_s:
            validate(json.load(f_d), json.load(f_s))


def fix_errorhandler(app: Flask) -> Flask:
    @app.errorhandler(400)
    @app.errorhandler(401)
    @app.errorhandler(403)
    @app.errorhandler(404)
    @app.errorhandler(500)
    def error_handler(error: HTTPException) -> Response:
        res_body: ErrorResponse = {
            "msg": error.description,  # type: ignore
            "status_code": error.code,  # type: ignore
        }
        response: Response = jsonify(res_body)
        response.status_code = error.code  # type: ignore
        return response

    @app.errorhandler(Exception)
    def error_handler_exception(exception: Exception) -> Response:
        current_app.logger.error(exception.args[0])
        current_app.logger.debug(format_exc())
        res_body: ErrorResponse = {
            "msg": "The server encountered an internal error and was "
                   "unable to complete your request.",
            "status_code": 500,
        }
        if current_app.config["TESTING"]:
            res_body["msg"] = format_exc()
        response: Response = jsonify(res_body)
        response.status_code = 500
        return response

    return app


def add_after_request(app: Flask) -> Flask:
    @app.after_request
    def after_request_func(response: Response) -> Response:
        response.headers["Access-Control-Allow-Origin"] = \
            os.environ.get("SAPPORO_ACCESS_CONTROL_ALLOW_ORIGIN",
                           DEFAULT_ACCESS_CONTROL_ALLOW_ORIGIN)
        response.headers["Access-Control-Allow-Headers"] = "Content-Type"

        return response

    return app


def create_app(params: Dict[str, Union[str, int, Path]]) -> Flask:
    app = Flask(__name__)
    app.register_blueprint(app_bp, url_prefix=params["url_prefix"])
    fix_errorhandler(app)
    add_after_request(app)
    app.config["RUN_DIR"] = params["run_dir"]
    app.config["GET_RUNS"] = params["get_runs"]
    app.config["WORKFLOW_ATTACHMENT"] = params["workflow_attachment"]
    app.config["REGISTERED_ONLY_MODE"] = params["registered_only_mode"]
    app.config["SERVICE_INFO"] = params["service_info"]
    app.config["EXECUTABLE_WORKFLOWS"] = params["executable_workflows"]
    app.config["RUN_SH"] = params["run_sh"]
    app.config["URL_PREFIX"] = params["url_prefix"]
    validate_json(app.config["SERVICE_INFO"],
                  app.config["EXECUTABLE_WORKFLOWS"])
    check_uniqueness_wf_name(app.config["EXECUTABLE_WORKFLOWS"])

    return app


def main() -> None:
    args: Namespace = parse_args(sys.argv[1:])
    params: Dict[str, Union[str, int, Path]] = handle_default_params(args)
    app: Flask = create_app(params)
    app.run(
        host=params["host"],  # type: ignore
        port=params["port"],  # type: ignore
        debug=params["debug"]  # type: ignore
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# coding: utf-8
from enum import Enum, auto
from sys import version_info
from typing import Any, Dict, List

if version_info.minor < 8:
    from typing_extensions import TypedDict
else:
    from typing import TypedDict  # type: ignore


class DefaultWorkflowEngineParameter(TypedDict):
    """
    Each workflow engine can present additional parameters that can be sent to
    the workflow engine. This message will list the default values, and their
    types for each workflow engine.

    name:
        The name of the parameter
    type:
        Describes the type of the parameter, e.g. float.
    default_value:
        The stringified version of the default parameter. e.g. "2.45".
    """
    name: str
    type: str
    default_value: str


class Log(TypedDict):
    """
    Log and other info

    name:
        The task or workflow name
    cmd:
        The command line that was executed
    start_time:
        When the command started executing, in ISO 8601 format
        "%Y-%m-%dT%H:%M:%SZ"
    end_time:
        When the command stopped executing (completed, failed, or cancelled),
        in ISO 8601 format "%Y-%m-%dT%H:%M:%SZ"
    stdout:
        A URL to retrieve standard output logs of the workflow run or task.
        This URL may change between status requests, or may not be available
        until the task or workflow has finished execution. Should be available
        using the same credentials used to access the WES endpoint.
    stderr:
        A URL to retrieve standard error logs of the workflow run or task.
        This URL may change between status requests, or may not be available
        until the task or workflow has finished execution. Should be available
        using the same credentials used to access the WES endpoint.
    exit_code:
        Exit code of the program
    """
    name: str
    cmd: List[str]
    start_time: str
    end_time: str
    stdout: str
    stderr: str
    exit_code: int


class State(Enum):
    """
    UNKNOWN:
        The state of the task is unknown. This provides a safe default for
        messages where this field is missing, for example, so that a missing
        field does not accidentally imply that the state is QUEUED.
    QUEUED:
        The task is queued.
    INITIALIZING:
        The task has been assigned to a worker and is currently preparing to
        run. For example, the worker may be turning on, downloading input
        files, etc.
    RUNNING:
        The task is running. Input files are downloaded and the first Executor
        has been started.
    PAUSED:
        The task is paused. An implementation may have the ability to pause a
        task, but this is not required.
    COMPLETE:
        The task has completed running. Executors have exited without error
        and output files have been successfully uploaded.
    EXECUTOR_ERROR:
        The task encountered an error in one of the Executor processes.
        Generally, this means that an Executor exited with a non-zero exit
        code.
    SYSTEM_ERROR:
        The task was stopped due to a system error, but not from an Executor,
        for example an upload failed due to network issues, the worker's ran
        out of disk space, etc.
    CANCELED:
        The task was canceled by the user.
    CANCELING:
        The task was canceled by the user, and is in the process of stopping.
    """
    UNKNOWN = auto()
    QUEUED = auto()
    INITIALIZING = auto()
    RUNNING = auto()
    PAUSED = auto()
    COMPLETE = auto()
    EXECUTOR_ERROR = auto()
    SYSTEM_ERROR = auto()
    CANCELED = auto()
    CANCELING = auto()


class WorkflowTypeVersion(TypedDict):
    """
    Available workflow types supported by a given instance of the service.

    workflow_type_version:
        an array of one or more acceptable types for the `workflow_type`
    """
    workflow_type_version: List[str]


class AttachedFile(TypedDict):
    """
    The file specified in `workflow_attachment`. It is expanded to the
    execution dir when the workflow is executed.

    file_name:
        File name. It can be also specified like `dir_name/file_name`.
    file_url:
        Remote file URL.
    """
    file_name: str
    file_url: str


class Workflow(TypedDict):
    """
    The information about the workflow provided by `service-info` in the mode
    of executing only the workflow registered by the administrator.

    workflow_name:
        The workflow name
    workflow_url:
        The workflow document url
    workflow_type:
        The workflow descriptor type, must be "CWL" or "WDL" currently (or
        another alternative supported by this WES instance)
    workflow_type_version:
        A map with keys as the workflow format type name (currently only CWL
        and WDL are used although a service may support others) and value is a
        workflow_type_version object which simply contains an array of one or
        more version strings
    workflow_attachment:
        The workflow_attachment array may be used to download files that are
        required to execute the workflow, including the primary workflow,
        tools imported by the workflow, other files referenced by the workflow,
        or files which are part of the input. The implementation should stage
        these files to a temporary directory and execute the workflow from
        there.
    """
    workflow_name: str
    workflow_url: str
    workflow_type: str
    workflow_type_version: str
    workflow_attachment: List[AttachedFile]


class ServiceInfo(TypedDict):
    """
    A message containing useful information about the running service,
    including supported versions and default settings.

    workflow_type_versions:
        A map with keys as the workflow format type name (currently only CWL
        and WDL are used although a service may support others) and value is a
        workflow_type_version object which simply contains an array of one or
        more version strings
    supported_wes_versions:
        The version(s) of the WES schema supported by this service
    supported_filesystem_protocols:
        The filesystem protocols supported by this service, currently these may
        include common protocols using the terms 'http', 'https', 'sftp', 's3',
        'gs', 'file', or 'synapse', but others are possible and the terms
        beyond these core protocols are currently not fixed. This section
        reports those protocols (either common or not) supported by this WES
        service.
    workflow_engine_versions:
        The engine(s) used by this WES service, key is engine name
        (e.g. Cromwell) and value is version
    default_workflow_engine_parameters:
        Each workflow engine can present additional parameters that can be
        sent to the workflow engine. This message will list the default values,
        and their types for each workflow engine.
    system_state_counts:
        The system statistics, key is the statistic, value is the count of
        runs in that state. See the State enum for the possible keys.
    auth_instructions_url:
        A web page URL with human-readable instructions on how to get an
        authorization token for use with a specific WES endpoint.
    contact_info_url:
        An email address URL (mailto:) or web page URL with contact
        information for the operator of a specific WES endpoint. Users of the
        endpoint should use this to report problems or security
        vulnerabilities.
    tags:
        A key-value map of arbitrary, extended metadata outside the scope of
        the above but useful to report back
    executable_workflows:
        List of workflows that can be executed
    """
    workflow_type_versions: Dict[str, WorkflowTypeVersion]
    supported_wes_versions: List[str]
    supported_filesystem_protocols: List[str]
    workflow_engine_versions: Dict[str, str]
    default_workflow_engine_parameters: List[DefaultWorkflowEngineParameter]
    system_state_counts: Dict[State, int]
    auth_instructions_url: str
    contact_info_url: str
    tags: Dict[str, str]
    executable_workflows: List[Workflow]


class RunStatus(TypedDict):
    """
    Small description of a workflow run, returned by server during listing
    """
    run_id: str
    state: State


class RunListResponse(TypedDict):
    """
    The service will return a RunListResponse when receiving a successful
    RunListRequest.

    runs:
        A list of workflow runs that the service has executed or is executing.
        The list is filtered to only include runs that the caller has
        permission to see.
    next_page_token:
        A token which may be supplied as `page_token` in workflow run list
        request to get the next page of results. An empty string indicates
        there are no more items to return.
    """
    runs: List[RunStatus]
    next_page_token: str


class RunRequest(TypedDict):
    """
    To execute a workflow, send a run request including all the details needed
    to begin downloading and executing a given workflow.

    workflow_params:
        The workflow run parameterizations (JSON encoded), including input and
        output file locations
    workflow_type:
        The workflow descriptor type, must be "CWL" or "WDL" currently (or
        another alternative supported by this WES instance)
    workflow_type_version:
        The workflow descriptor type version, must be one supported by this
        WES instance
    tags:
        A key-value map of arbitrary metadata outside the scope of
        `workflow_params` but useful to track with this run request
    workflow_engine_name:
        Specify the name of the workflow engine to run workflow.
    workflow_engine_parameters:
        Additional parameters can be sent to the workflow engine using this
        field. Default values for these parameters can be obtained using the
        ServiceInfo endpoint.
    workflow_url:
        The workflow CWL or WDL document. When `workflow_attachments` is used
        to attach files, the `workflow_url` may be a relative path to one of
        the attachments.
    workflow_name:
        The `workflow_name` is only used when SAPPORO is in the mode to
        execute only registered workflows. In the original EWS, it is
        OPTIONAL. To see the registered workflows, use `GET /service-info`.
    workflow_attachment:
        The workflow_attachment array may be used to upload files that are
        required to execute the workflow, including the primary workflow,
        tools imported by the workflow, other files referenced by the workflow,
        or files which are part of the input. The implementation should stage
        these files to a temporary directory and execute the workflow from
        there. These parts must have a Content-Disposition header with a
        "filename" provided for each part. Filenames may include
        subdirectories, but must not include references to parent directories
        with '..' -- implementations should guard against maliciously
        constructed filenames.
    """
    workflow_params: str
    workflow_type: str
    workflow_type_version: str
    tags: str
    workflow_engine_name: str
    workflow_engine_parameters: str
    workflow_url: str
    workflow_name: str
    workflow_attachment: List[AttachedFile]


class RunLog(TypedDict):
    """
    run_id:
      workflow run ID
    request:
      The original request message used to initiate this execution.
    state:
      The state of the run e.g. RUNNING (see State)
    run_log:
      The logs, and other key info like timing and exit code, for the overall
      run of this workflow.
    task_logs:
      The logs, and other key info like timing and exit code, for each step in
      the workflow run.
    outputs:
      The outputs from the workflow run.
    """
    run_id: str
    request: RunRequest
    state: State
    run_log: Log
    task_logs: List[Log]
    outputs: Dict[Any, Any]


class RunId(TypedDict):
    """
    workflow run ID
    """
    run_id: str


class ErrorResponse(TypedDict):
    """
    An object that can optionally include information about the error.

    msg:
        A detailed error message.
    status_code:
        The integer representing the HTTP status code (e.g. 200, 404).
    """
    msg: str
    status_code: int
